import trio


async def function_name():
    with (
        open("veryverylongfilenamesoshedsplitsthisintotwolines") as _,
        trio.fail_after(10),
    ):
        pass

    with (
        trio.fail_after(5),
        open("veryverylongfilenamesoshedsplitsthisintotwolines") as _,
        trio.move_on_after(5),
    ):
        pass



"""Tests for flake8-trio package metadata."""
import re
from pathlib import Path
from typing import NamedTuple, Optional

import flake8_trio


class Version(NamedTuple):
    year: int
    month: int
    day: int

    @classmethod
    def from_string(cls, string):
        return cls(*map(int, string.split(".")))


def get_latest_release() -> Optional[Version]:
    valid_pattern = re.compile(r"^## (\d\d\.\d?\d\.\d?\d)$")
    with open(Path(__file__).parent.parent / "CHANGELOG.md") as f:
        for aline in f.readlines():
            version_match = valid_pattern.match(aline)
            if version_match:
                return Version.from_string(version_match.group(1))

    return None


def test_last_release_against_changelog():
    """Ensure we have the latest version covered in CHANGELOG.md"""
    assert get_latest_release() == Version.from_string(flake8_trio.__version__)



import ast
import os
import site
import sys
import unittest
from pathlib import Path

import pytest
from hypothesis import HealthCheck, given, settings
from hypothesmith import from_grammar, from_node

from flake8_trio import TRIO100, TRIO101, Error, Plugin, Visitor, make_error


class Flake8TrioTestCase(unittest.TestCase):
    def assert_expected_errors(self, test_file: str, *expected: Error) -> None:
        filename = Path(__file__).absolute().parent / test_file
        plugin = Plugin.from_filename(str(filename))
        errors = tuple(plugin.run())
        self.assertEqual(errors, expected)

    def test_tree(self):
        plugin = Plugin(ast.parse(""))
        errors = list(plugin.run())
        self.assertEqual(errors, [])

    def test_trio100(self):
        self.assert_expected_errors(
            "trio100.py",
            make_error(TRIO100, 3, 5, "trio.move_on_after"),
            make_error(TRIO100, 23, 9, "trio.fail_after"),
        )

    @unittest.skipIf(sys.version_info < (3, 9), "requires 3.9+")
    def test_trio100_py39(self):
        self.assert_expected_errors(
            "trio100_py39.py",
            make_error(TRIO100, 7, 8, "trio.fail_after"),
            make_error(TRIO100, 12, 8, "trio.fail_after"),
            make_error(TRIO100, 14, 8, "trio.move_on_after"),
        )

    def test_trio101(self):
        self.maxDiff = None
        self.assert_expected_errors(
            "trio101.py",
            make_error(TRIO101, 10, 8),
            make_error(TRIO101, 15, 8),
            make_error(TRIO101, 27, 8),
            make_error(TRIO101, 38, 8),
        )


@pytest.mark.fuzz
class TestFuzz(unittest.TestCase):
    @settings(max_examples=1_000, suppress_health_check=[HealthCheck.too_slow])
    @given((from_grammar() | from_node()).map(ast.parse))
    def test_does_not_crash_on_any_valid_code(self, syntax_tree: ast.AST):
        # Given any syntatically-valid source code, the checker should
        # not crash.  This tests doesn't check that we do the *right* thing,
        # just that we don't crash on valid-if-poorly-styled code!
        Visitor().visit(syntax_tree)

    @staticmethod
    def _iter_python_files():
        # Because the generator isn't perfect, we'll also test on all the code
        # we can easily find in our current Python environment - this includes
        # the standard library, and all installed packages.
        for base in sorted(set(site.PREFIXES)):
            for dirname, _, files in os.walk(base):
                for f in files:
                    if f.endswith(".py"):
                        yield Path(dirname) / f

    def test_does_not_crash_on_site_code(self):
        for path in self._iter_python_files():
            try:
                Plugin.from_filename(str(path)).run()
            except Exception as err:
                raise AssertionError(f"Failed on {path}") from err



"""A highly opinionated flake8 plugin for Trio-related problems.

This can include anything from outright bugs, to pointless/dead code,
to likely performance issues, to minor points of idiom that might signal
a misunderstanding.

It may well be too noisy for anyone with different opinions, that's OK.

Pairs well with flake8-async and flake8-bugbear.
"""

import ast
import tokenize
from typing import Any, Generator, List, Optional, Set, Tuple, Type, Union

# CalVer: YY.month.patch, e.g. first release of July 2022 == "22.7.1"
__version__ = "22.7.2"


Error = Tuple[int, int, str, Type[Any]]
checkpoint_node_types = (ast.Await, ast.AsyncFor, ast.AsyncWith)
cancel_scope_names = (
    "fail_after",
    "fail_at",
    "move_on_after",
    "move_at",
    "CancelScope",
)
context_manager_names = (
    "contextmanager",
    "asynccontextmanager",
)


def make_error(error: str, lineno: int, col: int, *args: Any, **kwargs: Any) -> Error:
    return (lineno, col, error.format(*args, **kwargs), type(Plugin))


def is_trio_call(node: ast.AST, *names: str) -> Optional[str]:
    if (
        isinstance(node, ast.Call)
        and isinstance(node.func, ast.Attribute)
        and isinstance(node.func.value, ast.Name)
        and node.func.value.id == "trio"
        and node.func.attr in names
    ):
        return "trio." + node.func.attr
    return None


class Visitor(ast.NodeVisitor):
    def __init__(self) -> None:
        super().__init__()
        self.problems: List[Error] = []
        self.safe_yields: Set[ast.Yield] = set()
        self._yield_is_error = False
        self._context_manager = False

    def visit_generic_with(self, node: Union[ast.With, ast.AsyncWith]):
        self.check_for_trio100(node)

        outer = self._yield_is_error
        if not self._context_manager and any(
            is_trio_call(item, "open_nursery", *cancel_scope_names)
            for item in (i.context_expr for i in node.items)
        ):
            self._yield_is_error = True

        self.generic_visit(node)
        self._yield_is_error = outer

    def visit_With(self, node: ast.With) -> None:
        self.visit_generic_with(node)

    def visit_AsyncWith(self, node: ast.AsyncWith) -> None:
        self.visit_generic_with(node)

    def visit_generic_FunctionDef(
        self, node: Union[ast.FunctionDef, ast.AsyncFunctionDef]
    ):
        outer_cm = self._context_manager
        outer_yie = self._yield_is_error
        self._yield_is_error = False
        if any(
            (isinstance(d, ast.Name) and d.id in context_manager_names)
            or (isinstance(d, ast.Attribute) and d.attr in context_manager_names)
            for d in node.decorator_list
        ):
            self._context_manager = True
        self.generic_visit(node)
        self._context_manager = outer_cm
        self._yield_is_error = outer_yie

    def visit_FunctionDef(self, node: ast.FunctionDef) -> None:
        self.visit_generic_FunctionDef(node)

    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef) -> None:
        self.visit_generic_FunctionDef(node)

    def visit_Yield(self, node: ast.Yield) -> None:
        if self._yield_is_error:
            self.problems.append(make_error(TRIO101, node.lineno, node.col_offset))

        self.generic_visit(node)

    def check_for_trio100(self, node: Union[ast.With, ast.AsyncWith]) -> None:
        # Context manager with no `await` call within
        for item in (i.context_expr for i in node.items):
            call = is_trio_call(item, *cancel_scope_names)
            if call and not any(
                isinstance(x, checkpoint_node_types) for x in ast.walk(node)
            ):
                self.problems.append(
                    make_error(TRIO100, item.lineno, item.col_offset, call)
                )


class Plugin:
    name = __name__
    version = __version__

    def __init__(self, tree: ast.AST) -> None:
        self._tree = tree

    @classmethod
    def from_filename(cls, filename: str) -> "Plugin":
        with tokenize.open(filename) as f:
            source = f.read()
        return cls(ast.parse(source))

    def run(self) -> Generator[Tuple[int, int, str, Type[Any]], None, None]:
        visitor = Visitor()
        visitor.visit(self._tree)
        yield from visitor.problems


TRIO100 = "TRIO100: {} context contains no checkpoints, add `await trio.sleep(0)`"
TRIO101 = "TRIO101: yield inside a nursery or cancel scope is only safe when implementing a context manager - otherwise, it breaks exception handling"
